/** dslconf - Simple DSL modem configuration protocol. */
#include <poll.h>
#include <pthread.h>
#include <semaphore.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <arpa/inet.h>

#include "simplecom.h"
#include "dslconf.h"

#define max(a, b)	((a) > (b) ? (a) : (b))

#if defined(DEBUG)
#define dprint(...) { \
	fprintf(stderr, "simplecom: %s(): ", __func__); \
	fprintf(stderr, __VA_ARGS__); \
	fprintf(stderr, "\n"); \
}
#else
#define dprint(...)	/* nothing */
#endif

/** Check the size of a given dslconf packet.
 * @param[in] pkt Pointer to the buffer containing the packet to check.
 * @param[in] pktlen Length of buffer backing @pkt.
 * @return Zero on success, non-zero on error. */
static int check_dslconf_packetsize(struct dslconf_pkt *pkt, size_t pktlen)
{
	if (pktlen < sizeof(struct dslconf_hdr)) {
		printf("size_constraint: packet is too small to contain the dslconf header\n");
		return 1;
	}
	if (pktlen < sizeof(struct dslconf_hdr) + ntohl(pkt->hdr.dlen)) {
		printf("size_constraint: packet is too small to contain specified amount of data\n");
		return 1;
	}
	return 0;
}

/** Call recv_packet and make sure it's a sane dslconf packet
 * There is a specialty for dslconf here: as it consists mainly of ASCII payloads,
 * append a zero byte to the received packet so standard string functions can be
 * used to print the payload.
 * @param[in] iface Interface to receive the packet(s) on.
 * @param[out] buf Buffer pointing to the full dslconf packet after returning to caller.
 * @param[out] buflen Length of @buf.
 * @return @buflen on success, negative value on error or zero if packet wasn't
 * interesting (see recv_packet() for details). */
ssize_t recv_dslconf_packet(struct interface iface, void **buf, __u32 *buflen)
{
	ssize_t rc;

	/* if recv_packet either fails or has something unrelevant, return directly */
	if ((rc = recv_packet(iface, buf, buflen)) <= 0)
		return rc;

	/* check packetsize, return -2 like recv_packet does for incomplete packets */
	if (check_dslconf_packetsize((struct dslconf_pkt *)(*buf), *buflen)) {
		printf("got an incomplete dslconf packet - broken sender?\n");
		return -2;
	}

	/* special for printing ascii content: extend buffer by binary zero */
	(*buf) = realloc((*buf), *buflen + 1);
	((char *)(*buf))[*buflen] = '\0';

	return rc;
}

/** Allocate a dslconf packet
 * As with receive_dslconf_packet, a zero byte at the end of the returned buffer
 * is additionally allocated so payload can be treated as standard string. To
 * prevent sending this byte, the header's 'dlen' field does not account for it.
 * @param[in] type Packet type (header field).
 * @param[in] flags Packet flags (header field).
 * @param[in] payload Payload to copy into packet if not NULL.
 * @param[in] payload_len Length of payload to allocate for.
 * @return Allocated packet or NULL on error. */
struct dslconf_pkt *alloc_dslconf_pkt(__u8 type, __u8 flags,
                                      const void *payload, size_t payload_len)
{
	struct dslconf_pkt *pkt;
	size_t pktlen;

	/* packet consists of header, payload and additional zero byte */
	pktlen = sizeof(struct dslconf_hdr) + payload_len + 1;

	/* ethernet requires minimum packet size of 60 bytes */
	pktlen = max(pktlen, 60);

	pkt = malloc(pktlen);
	if (pkt == NULL)
		return NULL;

	pkt->hdr.type = type;
	pkt->hdr.flags = flags;
	pkt->hdr.dlen = htonl(payload_len);

	if (payload)
		memcpy(pkt->data, payload, payload_len);
	memset(pkt->data + payload_len, 0,
	       pktlen - sizeof(struct dslconf_hdr) - payload_len);

	return pkt;
}

/** Do a complete client transaction (i.e. send a packet and receive the reply).
 * @param[in] iface Interface to use for sending and receiving.
 * @param[in] req Packet to send out.
 * @param[out] rep Pointer to the reply packet's address, if any.
 * @return -1 if an error occurred, 1 on timeout, zero otherwise. */
int do_transaction(struct interface iface, struct dslconf_pkt *req, struct dslconf_pkt **rep)
{
	struct dslconf_pkt *rpkt;
	struct pollfd pfd;
	int err, try = 0;
	__u32 rbuflen;
	void *rbuf;

	pfd.fd = iface.sockfd;
	pfd.events = POLLIN;

	if (send_packet(iface, req, sizeof(*req) + ntohl(req->hdr.dlen))) {
		printf("send_packet() returned not good\n");
		return -1;
	}
	dprint("sent packet with data: '%s'\n", req->data);

poll_again:
	if ((err = poll(&pfd, 1, 1000)) == -1) {
		perror("poll()");
		return -1;
	}
	if (!err) {
		if (try++ > 5) {
			return 1;
		}
		goto poll_again;
	}
	if ((err = recv_dslconf_packet(iface, &rbuf, &rbuflen)) <= 0) {
		/* receive packets until something valid turns up */
		goto poll_again;
	}

	rpkt = (struct dslconf_pkt *)rbuf;

	if (rpkt->hdr.type != req->hdr.type) {
		goto poll_again;
	}
	if (!(rpkt->hdr.flags & (FLAG_ACK | FLAG_NACK))) {
		goto poll_again;
	}
	(*rep) = rpkt;
	return 0;
}

static int listener_callback_dummy(struct interface iface,
                                   struct dslconf_pkt *pkt)
{
	printf("packet received on interface %s with type %d and nobody cares!\n",
			iface.name, pkt->hdr.type);
	return 1;
}

/** Listen for requests, run callback as appropriate.
 * @param[in] iface Interface to listen on.
 * @param[in] lcb Array of callback definitions to lookup the handler from.
 * @return Doesn't return. */
void listener(struct interface iface, const struct listener_callback *lcb, sem_t *sem)
{
	void *data;
	__u32 datalen;
	struct dslconf_pkt *dslpkt;
	/* type is __u8, so 256 entries suffice */
	int (*callback[256])(struct interface iface, struct dslconf_pkt *pkt);
	int idx;

	for (idx = 0; idx < 256; idx++)
		callback[idx] = &listener_callback_dummy;

	for (; lcb->callback; lcb++)
		callback[lcb->type] = lcb->callback;

	while(1) {
		while (recv_dslconf_packet(iface, &data, &datalen) <= 0) {
			/* receive until something valid turns up */
			continue;
		}
		dslpkt = (struct dslconf_pkt *)data;

		if (!(dslpkt->hdr.flags & FLAG_REQ)) { /* not a request */
			free(data);
			continue;
		}
		dprint("listener got a packet:\n\tlength %zd\n\toverhead %zd\n\ttype %d\n\t"
		       "flags %d\n\tdlen %u\n\tdata '%s'\n", datalen, sizeof(struct dslconf_hdr),
		       dslpkt->hdr.type, dslpkt->hdr.flags, ntohl(dslpkt->hdr.dlen), dslpkt->data);

		if (sem)
			sem_wait(sem);
		(*callback[dslpkt->hdr.type])(iface, dslpkt);
		if (sem)
			sem_post(sem);

		free(data);
	}
}

/** Send a reply.
 * Just send out a buffer of data with given length.
 *
 * @param[in] iface Interface to send the reply out on.
 * @param[in] type Request type to send a reply for.
 * @param[in] flags Server flags to set (FLAG_{ACK,NACK})
 * @param[in] data Data to return.
 * @param[in] datalen Length of data.
 * @return Zero on success, non-zero otherwise. */
int send_reply_pkt(struct interface iface, __u8 type, __u8 flags,
		const void *data, __u32 datalen)
{
	struct dslconf_pkt *rpkt;
	__u32 pktlen, rc;

	pktlen = sizeof(struct dslconf_hdr) + datalen;

	rpkt = malloc(pktlen);
	memset(rpkt, 0, pktlen);

	rpkt->hdr.type = type;
	rpkt->hdr.flags = flags;
	rpkt->hdr.dlen = htonl(datalen);
	memcpy(rpkt->data, data, datalen);

	rc = send_packet(iface, rpkt, pktlen);

	free(rpkt);
	return rc;
}
