/** simplecom - A simple communication protocol for two peers in an ethernet segment. */
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <net/ethernet.h>
#include <netpacket/packet.h>
#include <sys/ioctl.h>

/* WAR stupid broken montavista headers */
#ifdef TARGET_AR7
#  define ETH_CRC_LEN ETHER_CRC_LEN
#endif

#include "simplecom.h"

#define MAX(a, b) ((a) > (b) ? (a) : (b))

#if defined(DEBUG)
#define dprint(...) { \
	fprintf(stderr, "simplecom: %s(): ", __func__); \
	fprintf(stderr, __VA_ARGS__); \
	fprintf(stderr, "\n"); \
}
#else
#define dprint(...)	/* nothing */
#endif

/** Initialise an interface description.
 * @param[in] ifname Interface name to create a description for.
 * @param[in] proto Next layer protocol type.
 * @return Newly allocated interface description, or NULL on error. */
struct interface *init_interface(const char *ifname, __u8 proto)
{
	struct interface *retval;
	struct ifreq ifr;

	if (!ifname)
		return NULL;

	retval = malloc(sizeof(struct interface));
	memset(retval, 0, sizeof(struct interface));

	snprintf(retval->name, IFNAMSIZ, "%s", ifname);
	retval->index = if_nametoindex(ifname);
	retval->proto = proto;

	/* open a socket */
	retval->sockfd = socket(AF_PACKET, SOCK_DGRAM, htons(ETH_P_SCOM));
	if (retval->sockfd < 0) {
		perror("socket()");
		goto err_out;
	}

	/* retrieve interface's hwaddr */
	strncpy(ifr.ifr_name, ifname, IFNAMSIZ);
	if (ioctl(retval->sockfd, SIOCGIFHWADDR, &ifr)) {
		perror("ioctl(SIOCGIFHWADDR)");
		goto err_out;
	}
	memcpy(retval->hwaddr, ifr.ifr_hwaddr.sa_data, ETH_ALEN);
	dprint("got mac address %02x:%02x:%02x:%02x:%02x:%02x for device %s\n",
	       (unsigned char)retval->hwaddr[0], (unsigned char)retval->hwaddr[1],
	       (unsigned char)retval->hwaddr[2], (unsigned char)retval->hwaddr[3],
	       (unsigned char)retval->hwaddr[4], (unsigned char)retval->hwaddr[5], ifname);

	/* retrieve the MTU */
	if (ioctl(retval->sockfd, SIOCGIFMTU, &ifr)) {
		perror("ioctl(SIOCGIFMTU)");
		goto err_out;
	}
	retval->mtu = (ifr.ifr_mtu > MAX_MTU ? MAX_MTU : ifr.ifr_mtu);

	/* prepare a struct sockaddr for the following bind() and all later sending tasks */
	retval->send_sa.sll_family = AF_PACKET;
#ifdef TARGET_FUSIV
	/* For whatever reason, binding to ETH_P_SCOM makes the socket receive nothing
	 * when the interface it's bound to is part of a bridge. Happens only on Fusiv. */
	retval->send_sa.sll_protocol = htons(ETH_P_ALL);
#else
	retval->send_sa.sll_protocol = htons(ETH_P_SCOM);
#endif
	retval->send_sa.sll_halen = ETH_ALEN;
	retval->send_sa.sll_ifindex = retval->index;
	retval->send_sa.sll_pkttype = PACKET_BROADCAST;
	memcpy(retval->send_sa.sll_addr, retval->hwaddr, ETH_ALEN);

	/* bind socket to this interface */
	if (bind(retval->sockfd, (struct sockaddr *)&retval->send_sa, sizeof(retval->send_sa))) {
		perror("bind()");
		goto err_out;
	}
#ifdef TARGET_FUSIV
	/* reset the protocol field again for use when sending */
	retval->send_sa.sll_protocol = htons(ETH_P_SCOM);
#endif
	/* prepare for sending to the broadcast address */
	memcpy(retval->send_sa.sll_addr, "\xff\xff\xff\xff\xff\xff", ETH_ALEN);

	/* Prepare an ethernet header for use when sending packets.
	 * The destination address is always the layer2 broadcast address (all one's). */
	memcpy(retval->eh.ether_shost, retval->hwaddr, ETH_ALEN);
	memcpy(retval->eh.ether_dhost, "\xff\xff\xff\xff\xff\xff", ETH_ALEN);
	retval->eh.ether_type = htons(ETH_P_SCOM);

	return retval;
err_out:
	if (retval->sockfd)
		close(retval->sockfd);
	free(retval);
	return NULL;
}

/** Free data used by an interface description.
 * @param[in,out] iface Interface description to free. */
void exit_interface(struct interface *iface)
{
	close(iface->sockfd);
	free(iface);
}

/** Check the sise of a given buffer pointing to an scom packet.
 * @param[in] buf Buffer containing the packet.
 * @param[in] buflen Length of @buf.
 * @return Zero on success, non-zero otherwise. */
static int check_scom_packetsize(void *buf, __u16 buflen)
{
	if (buflen < sizeof(struct scom_hdr))
		return 1;
	if (buflen < sizeof(struct scom_hdr) +
	    ntohs(((struct scom_hdr *)buf)->pktlen))
		return 1;
	return 0;
}

/** Check if a sockaddr_ll matches for a received packet.
 * @param[in] sa Sockaddr_ll to match.
 * @param[in] iface Interface description to retrieve information from.
 * @return TRUE if packet is simplecom and did not origin from us,
 *	FALSE otherwise. */
static int match_sockaddr(struct sockaddr_ll sa, struct interface iface)
{
#ifdef TARGET_FUSIV
	/* since socket is not bound to that protocol on fusiv, it receives everything */
	if (sa.sll_protocol != htons(ETH_P_SCOM))
		return 0;
#endif
	/* must not origin from us */
	if (!memcmp(sa.sll_addr, iface.hwaddr, ETH_ALEN))
		return 0;
	return 1;
}

/** Send a single packet of matching size.
 * @param[in] iface Interface to send packet out on.
 * @param[in] buf Buffer containing the full ethernet payload.
 * @param[in] buflen Length of @buf.
 * @return Zero on success, negative value otherwise. */
static ssize_t send_single_packet(struct interface iface, void *buf, __u16 buflen)
{
	ssize_t rc;

	/* don't allow accidential fragmentation / packet cropping */
	assert(buflen <= iface.mtu);

	/* don't send undersized ethernet frames */
	assert(buflen >= ETHERMIN);

	dprint("calling sendto with a buffer of %d bytes\n", buflen);
	rc = sendto(iface.sockfd, buf, buflen, 0,
			(struct sockaddr *)&iface.send_sa, sizeof(iface.send_sa));
	if (rc == -1)
		perror("sendto()");

	return rc - buflen;
}

/** Send a packet with optional fragmentation.
 * This routine splits the given payload into chunks of appropriate size if
 * necessary, adds an SCOM header to each and passes them to
 * send_single_packet() for the actual delivery.
 * @param[in] iface Interface used for sending.
 * @param[in] buf Buffer containing the full scom payload.
 * @param[in] buflen Length of @buf.
 * @return Zero on success, negative value otherwise. */
ssize_t send_packet(struct interface iface, void *buf, __u32 buflen)
{
	__u32 offset;
	__u16 mss, pktsize, datalen;
	__u8 *pktbuf;
	struct scom_hdr *shp;
	ssize_t rc = 0;

	mss = iface.mtu - sizeof(struct scom_hdr);

	pktbuf = malloc(iface.mtu);
	shp = (struct scom_hdr *)pktbuf;

	if (buflen > mss) {
		dprint("fragmentation needed\n");
	}

	for (offset = 0; offset < buflen; offset += mss) {
		datalen = (offset + mss) <= buflen ? mss : buflen - offset;
		pktsize = MAX(datalen, ETHERMIN - sizeof(*shp));

		/* fill in scom_hdr values */
		shp->seqno = htonl(iface.seqno++);
		shp->proto = iface.proto;
		/* if it's not the first packet, it must be a fragment at least */
		shp->flags = (offset > 0) ? SCFL_FRAG : SCFL_NONE;
		/* if there's more about to come, it's fragment with followers */
		if ((offset + mss) < buflen)
			shp->flags = SCFL_FRAG | SCFL_MF;
		shp->pktlen = htons(datalen);

		memcpy(pktbuf + sizeof(*shp), (__u8 *)buf + offset, datalen);
		if (pktsize > datalen)
			memset(pktbuf + sizeof(*shp) + datalen, 0, pktsize - datalen);

		if ((rc = send_single_packet(iface, pktbuf, pktsize + sizeof(*shp)))) {
			printf("failed to send packet\n");
			break;
		}
	}

	free(pktbuf);
	return rc;
}

/** Receive a single packet from the line.
 * This routine inspects the ethernet header to determine if the packet is
 * interesting (i.e. sent by the peer) or not, returning an appropriate value.
 * @param[in] iface Interface to receive the packet from.
 * @param[out] buf Buffer pointing to the ethernet payload after return.
 * @param[out] buflen Length of buffer @buf points to.
 * @return Negative value if an error occurred,	zero if packet is not
 *	interesting (e.g., was sent by ourself), positive value on success. */
ssize_t recv_single_packet(struct interface iface, void **buf, __u16 *buflen)
{
	struct sockaddr_ll sa;
	socklen_t salen = sizeof(sa);
	__u16 pktbuflen;
	__u8 *pktbuf;
	ssize_t rc;

	pktbuflen = iface.mtu;
	pktbuf = malloc(pktbuflen);
	if (!pktbuf) {
		printf("out of memory\n");
		return -1;
	}
	memset(pktbuf, 0, pktbuflen);

	rc = recvfrom(iface.sockfd, pktbuf, pktbuflen, 0,
			(struct sockaddr *)&sa, &salen);
	if (rc == -1) {
		perror("recvfrom()");
		goto out_err;
	}
	if (!match_sockaddr(sa, iface)) {
		rc = 0;	/* packet not interesting */
		goto out_err;
	}

	if (check_scom_packetsize(pktbuf, rc)) {
		printf("got an incomplete packet - broken medium?\n");
		rc = -2;
		goto out_err;
	}
	if (((struct scom_hdr *)pktbuf)->proto != iface.proto) {
		rc = 0; /* nothing for us */
		goto out_err;
	}

	if (buf)
		*buf = pktbuf;
	if (buflen)
		*buflen = rc;
	return rc;
out_err:
	if (pktbuf)
		free(pktbuf);
	return rc;

}

/** Receive a full (dslconf) packet from the line.
 * When the first received packet is a fragment, this routine starts collecting
 * further packets until all fragments have been received. In case of an error
 * (e.g. an incoming, non-matching packet) the whole collection is being
 * dropped.
 * @param[in] iface Interface to receive the packet(s) on.
 * @param[out] buf Buffer pointing to the full dslconf packet after returning to caller.
 * @param[out] buflen Length of @buf.
 * @return @buflen on success, negative value on error or zero if packet wasn't
 * interesting (see recv_single_packet() for details). */
ssize_t recv_packet(struct interface iface, void **buf, __u32 *buflen)
{
	__u8 *savebuf;
	void *tmpbuf = NULL;
	__u32 savebuflen, seqno;
	__u8 mf;
	__u16 tmpbuflen;
	struct scom_hdr *shp;
	ssize_t rc;

	rc = recv_single_packet(iface, &tmpbuf, &tmpbuflen);

	dprint("receive_single_packet returned %d\n", rc);
	if (rc <= 0)
		return rc;

	shp = (struct scom_hdr *)tmpbuf;

	if (!(shp->flags & (SCFL_FRAG | SCFL_MF))) {
		tmpbuflen -= sizeof(*shp);
		memmove(tmpbuf, (__u8 *)tmpbuf + sizeof(*shp), tmpbuflen);
		*buf = tmpbuf;
		*buflen = tmpbuflen;
		return rc;
	}
	dprint("fragment received\n");

	seqno = ntohl(shp->seqno);

	savebuf = malloc(tmpbuflen - sizeof(*shp));
	savebuflen = tmpbuflen - sizeof(*shp);
	memcpy(savebuf, (__u8 *)tmpbuf + sizeof(*shp), tmpbuflen - sizeof(*shp));
	free(tmpbuf);

	do {
		rc = recv_single_packet(iface, &tmpbuf, &tmpbuflen);
		if (rc <= 0) {
			printf("receive single packet failed for missing fragments\n");
			break;
		}

		shp = (struct scom_hdr *)tmpbuf;
		if (!(shp->flags & SCFL_FRAG)) {
			printf("non-fragment received when one was expected, ignoring whole transfer\n");
			rc = -rc;
			break;
		}
		if (++seqno != ntohl(shp->seqno)) {
			printf("non-monotonic seqno found in fragment (got %d, expected %d)\n", ntohl(shp->seqno), seqno);
			rc = -rc;
			break;
		}
		mf = shp->flags & SCFL_MF;
		dprint("%s fragment received (seqno %d)\n", mf ? "next" : "last", ntohl(shp->seqno));

		savebuf = realloc(savebuf, savebuflen + tmpbuflen - sizeof(*shp));
		memcpy(savebuf + savebuflen, (__u8 *)tmpbuf + sizeof(*shp),
				tmpbuflen - sizeof(*shp));
		savebuflen += tmpbuflen - sizeof(*shp);
		free(tmpbuf);
		tmpbuf = NULL;
	} while (mf);

	if (rc <= 0) {
		if (tmpbuf)
			free(tmpbuf);
		free(savebuf);
		return rc;
	}
	*buf = savebuf;
	*buflen = savebuflen;
	return savebuflen;
}

