/*
 * Copyright Cellient Corp. (C) 2010
 * Jayden Kim <superbolt@cellient.com>
 */

char *version="1.0.0";

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <signal.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <termios.h>
#include <ctype.h>
#include <getopt.h>
#include <syslog.h>

#include "fcs.h"


int             dm_quiet = 0;
char            dm_port[32];
char            dm_file[32];
struct termios  dm_tios;

#define         MDM_TX_RETRY        2
#define         MDM_RX_RETRY        4

#define                    LINE_DIM 1024
unsigned char   dm_txbuf [ LINE_DIM ];
unsigned char   dm_rxbuf [ LINE_DIM ];
unsigned int    dm_txlen = 0;
unsigned int    dm_rxlen = 0;

static struct option long_options[] = {
    {"cmd",       required_argument,  0,    'c'},
    {"port",      required_argument,  0,    'p'},
    {"file",      required_argument,  0,    'f'},
    {"help",      no_argument,        0,    'h'},
    {"version",   no_argument,        0,    'v'},
    {0,                     0,        0,     0}
};

void dm_print_version()
{
    printf("\n");
    printf(" * mdm: Mobile Diagnostic Monitor for USB modem\n");
    printf(" * Version %s (C) Jayden Kim 2010\n", version);
    printf("\n");
}

void dm_print_usage()
{
    printf ("Usage: mdm [-hvqs] [-p port] [-c dmcmd]\n\n");
    printf (" -h, --help        this help\n");
    printf (" -v, --version     print version number and exit\n");
    printf (" -q, --quiet       write atcop result to default file\n");
    printf (" -s, --show        do show progress or error messages\n");
    printf (" -d, --command     DM command to be send to modem\n");
    printf (" -p, --port        DM port\n");
    printf (" -f, --file        file name for saving result\n");
}

static int hex2num(char c)
{
    if( c >= '0' && c <= '9' )
        return c - '0';
    if( c >= 'a' && c <= 'f' )
        return c - 'a' + 10;
    if( c >= 'A' && c <= 'F' )
        return c - 'A' + 10;
    return -1;
}

static int hex2byte(const char *hex)
{
    int a, b;
    a = hex2num(*hex++);
    if( a < 0 )
        return -1;
    b = hex2num(*hex++);
    if( b < 0 )
        return -1;
    return(a << 4) | b;
}

static unsigned int xstoi(char *hexstr)
{
    int           n = 0;      /* position in string               */
    int           m = 0;      /* position in digit[] to shift     */
    int           length;     /* loop index                       */
    int           digit[9];   /* hold values to convert           */
    unsigned int  value = 0;  /* double word value of hex string  */

    while( n < 8 ) {
        if( hexstr[n]=='\0' )
            break;
        if( hexstr[n] > 0x29 && hexstr[n] < 0x40 )      /* if 0 to 9      */
            digit[n] = hexstr[n] & 0x0f;                /* convert to int */
        else if( hexstr[n] >='a' && hexstr[n] <= 'f' )  /* if a to f      */
            digit[n] = (hexstr[n] & 0x0f) + 9;          /* convert to int */
        else if( hexstr[n] >='A' && hexstr[n] <= 'F' )  /* if A to F      */
            digit[n] = (hexstr[n] & 0x0f) + 9;          /* convert to int */
        else
            break;

        n++;
    }

    length = n;
    m = n - 1;
    n = 0;
    while( n < length ) {
        /* digit[n] is value of hex digit at position n
        ** (m << 2) is the number of positions to shift
        ** OR the bits into return value
        */
        value = value | (digit[n] << (m << 2));
        m--;    /* adjust the position to set */
        n++;    /* next digit to process      */
    }

    return(value);
}

/* read a set number of lines of text from the port */
int read_lines(int fd, char* buf, int lines)
{
    int count;
    int line, row = 0;

    for( line = 0; lines < lines; line++ ) {
        for( ; buf[row] != '\n'; row++ ) {
            if( 0 >= read(fd, (buf + row), 1) )  // port closed for some reason
                return -count;
            else {
                putchar(buf[row]);
                count++;
            }
        }
    }
    return count;
}

void dm_make_pkt( unsigned char *buf, unsigned int *len )
{
    /* DIAG packet */
    unsigned short fcs;
    unsigned int   idx;
    unsigned int   siz;
    unsigned char  tmp[ LINE_DIM ] = { '\0',};

    siz = 0;
    for( idx = 0; idx < *len; idx += 2 ) {
        tmp[ siz++ ] = hex2byte( (char*)&buf[idx] );
    }

    /* nv read request */
    if( 0x26 == tmp[ 0 ] ) {
        for( idx = 0; idx < 130 /* data:128 + status:2 */; idx++ ) {
            tmp[ siz++ ] = 0x00;
        }
    }

    memcpy( buf, tmp, siz );
    fcs  = pppfcs(PPPINITFCS, tmp, siz);
    fcs ^= PPPINITFCS;

    buf[ siz++ ] = (fcs & 0x00FF);
    buf[ siz++ ] = ((fcs >>8) & 0x00FF);
    buf[ siz++ ] = 0x7E;
    buf[ siz   ] = '\0';
    *len = siz;
}

int dm_put_line( int fd, const unsigned char *buf, unsigned int len )
{
    int loop = len;
    while( loop ) {
        int wr = write( fd, (char*)&buf[len-loop], loop );
        if( wr < 0 ) {
            printf( "Could not write to modem (err:%d)\n", wr );
            return -1;
        }
        printf( "Success to send %d bytes\n", wr );
        loop -= wr;
        tcdrain( fd );
    }

    tcdrain( fd );
    return 0;
}

int dm_get_line( int fd, unsigned char *buf, unsigned int len )
{
    int rr;

    memset( buf, 0x00, len );

    rr = read( fd, (char*)buf, len );
    if( rr <= 0 ) {
        printf( "Could not read any bytes (err:%d)\n", rr );
        return 0;
    }

    printf( "Success to receive %d bytes\n", rr );
    return rr;
}

static unsigned int get_value( unsigned char *buf, int *idx, int siz )
{
    int           ii;
    unsigned int  value = 0;

    for( ii = siz - 1; ii >= 0; ii-- ) {
        value = (value << 8) | buf[ *idx + ii ];
    }

    *idx += siz;

    return value;
}

int dm_parse_state( unsigned char *buf, unsigned int len )
{
    int  index = 0;
    int  value = 0;
    char buffer[ LINE_DIM + 2 ] = "";

    value = buf[ index++ ]; /* command = 0xCC */
    value = buf[ index++ ]; /* length */

    value = buf[ index++ ];
    sprintf( buffer + strlen(buffer), "HDR MODE: %d\n", value );

    value = buf[ index++ ];
    sprintf( buffer + strlen(buffer), "SYS SRV MODE: %d\n", value );

    value = buf[ index++ ];
    sprintf( buffer + strlen(buffer), "1x RSSI: %d\n", value );

    value = buf[ index++ ];
    sprintf( buffer + strlen(buffer), "EVDO RSSI: %d\n", value );

    value = buf[ index++ ];
    sprintf( buffer + strlen(buffer), "ANT LEVEL: %d\n", value );

    value = buf[ index++ ];
    sprintf( buffer + strlen(buffer), "ROAM IND: %d\n", value );

    value = buf[ index++ ];
    sprintf( buffer + strlen(buffer), "PKT STATE: %d\n", value );

    value = buf[ index++ ];
    sprintf( buffer + strlen(buffer), "MIP ERR CODE: %d\n", value );

    value = buf[ index++ ];
    sprintf( buffer + strlen(buffer), "LIN OFF: %d\n", value );

    int buflen = strlen( buffer );
    if( buflen > LINE_DIM ) {
        printf( "small rxbuf size(%d) than %d\n", len, buflen );
        buflen = LINE_DIM;
        buffer[ buflen ] = '\0';
    }

    memcpy( (char*)buf, buffer, buflen );
    return buflen;
}

int dm_parse_status( unsigned char *buf, unsigned int len )
{
    int  index = 0;
    int  value = 0;
    char tmpbuf[ 12 ];
    char buffer[ LINE_DIM + 2 ] = "";

    // CMD_CODE                     1
    value = get_value( buf, &index, 1 );

    // RESERVED                     3
    value = get_value( buf, &index, 3 );

    // ESN                          4
    value = get_value( buf, &index, 4 );
    sprintf( buffer + strlen(buffer), "ESN: 0x%x\n", value );

    // RF_MODE                      2
    value = get_value( buf, &index, 2 );
    sprintf( buffer + strlen(buffer), "RF_MODE: %d\n", value );

    // MIN1 (Analog)                4
    value = get_value( buf, &index, 4 );
    sprintf( buffer + strlen(buffer), "MIN1 (Analog): %d\n", value );

    // MIN1 (CDMA)                  4
    value = get_value( buf, &index, 4 );
    sprintf( buffer + strlen(buffer), "MIN1 (CDMA): %d\n", value );

    // MIN2 (Analog)                2
    value = get_value( buf, &index, 2 );
    sprintf( buffer + strlen(buffer), "MIN2 (Analog): %d\n", value );

    // MIN2 (CDMA)                  2
    value = get_value( buf, &index, 2 );
    sprintf( buffer + strlen(buffer), "MIN2 (CDMA): %d\n", value );

    // RESERVED                     1
    value = get_value( buf, &index, 1 );

    // CDMA_RX_STATE                2
    value = get_value( buf, &index, 2 );
    sprintf( buffer + strlen(buffer), "CDMA_RX_STATE: %d\n", value );

    // CDMA_GOOD_FRAMES             1
    value = get_value( buf, &index, 1 );
    sprintf( buffer + strlen(buffer), "CDMA_GOOD_FRAMES: %d\n", value );

    // ANALOG_CORRECTED_FRAMES      2
    value = get_value( buf, &index, 2 );
    sprintf( buffer + strlen(buffer), "ANALOG_CORRECTED_FRAMES: %d\n",value );

    // ANALOG_BAD_FRAMES            2
    value = get_value( buf, &index, 2 );
    sprintf( buffer + strlen(buffer), "ANALOG_BAD_FRAMES: %d\n", value );

    // ANALOG_WORD_SYNCS            2
    value = get_value( buf, &index, 2 );
    sprintf( buffer + strlen(buffer), "ANALOG_WORD_SYNCS: %d\n", value );

    // ENTRY_REASON                 2
    value = get_value( buf, &index, 2 );
    sprintf( buffer + strlen(buffer), "ENTRY_REASON: %d\n", value );

    // CURRENT_CHAN                 2
    value = get_value( buf, &index, 2 );
    sprintf( buffer + strlen(buffer), "CURRENT_CHAN: %d\n", value );

    // CDMA_CODE_CHAN               1
    value = get_value( buf, &index, 1 );
    sprintf( buffer + strlen(buffer), "CDMA_CODE_CHAN: %d\n", value );

    // PILOT_BASE                   2
    value = get_value( buf, &index, 2 );
    sprintf( buffer + strlen(buffer), "PILOT_BASE: %d\n", value );

    // SID                          2
    value = get_value( buf, &index, 2 );
    sprintf( buffer + strlen(buffer), "SID: %d\n", value );

    // NID                          2
    value = get_value( buf, &index, 2 );
    sprintf( buffer + strlen(buffer), "NID: %d\n", value );

    // LOCAID                       2
    value = get_value( buf, &index, 2 );
    sprintf( buffer + strlen(buffer), "LOCAID: %d\n", value );

    // RSSI                         2
    value = get_value( buf, &index, 2 );
    sprintf( buffer + strlen(buffer), "RSSI: %d\n", value );

    // POWER                        1
    value = get_value( buf, &index, 1 );
    sprintf( buffer + strlen(buffer), "POWER: %d\n", value );

    int buflen = strlen( buffer );
    if( buflen > LINE_DIM ) {
        printf( "small rxbuf size(%d) than %d\n", len, buflen );
        buflen = LINE_DIM;
        buffer[ buflen ] = '\0';
    }

    memcpy( (char*)buf, buffer, buflen );
    return buflen;
}

int dm_parse_pilot( unsigned char *buf, unsigned int len )
{
    int  index = 0;
    int  value = 0;
    char buffer[ LINE_DIM + 2 ] = "";

    // CMD_CODE (64 / 0x40)         1
    value = get_value( buf, &index, 1 );

    // PILOT_INC                    2
    value = get_value( buf, &index, 2 );
    sprintf( buffer + strlen(buffer), "PILOT_INC: %d\n", value );

    // ACTIVE_CNT                   1
    value = get_value( buf, &index, 1 );
    sprintf( buffer + strlen(buffer), "ACTIVE_CNT: %d\n", value );

    // CANDIDATE_CNT                1
    value = get_value( buf, &index, 1 );
    sprintf( buffer + strlen(buffer), "CANDIDATE_CNT: %d\n", value );

    // NEIGHBOR_CNT                 1
    value = get_value( buf, &index, 1 );
    sprintf( buffer + strlen(buffer), "NEIGHBOR_CNT: %d\n", value );

    // PILOT_PN                     2
    value = get_value( buf, &index, 2 );
    sprintf( buffer + strlen(buffer), "PILOT_PN: %d\n", value );

    // PILOT_STRENGTH               2
    value = get_value( buf, &index, 2 );
    sprintf( buffer + strlen(buffer), "PILOT_STRENGTH: %d\n", value );

    int buflen = strlen( buffer );
    if( buflen > LINE_DIM ) {
        printf( "small rxbuf size(%d) than %d\n", len, buflen );
        buflen = LINE_DIM;
        buffer[ buflen ] = '\0';
    }

    memcpy( (char*)buf, buffer, buflen );
    return buflen;
}

int dm_parse_ruim_pin( unsigned char *buf, unsigned int len )
{
    int  index = 0;
    int  value = 0;
    char buffer[ LINE_DIM + 2 ] = "";

    // CMD_CODE (0x8D)              1
    value = get_value( buf, &index, 1 );

    // SUB_CMD                      1
    value = get_value( buf, &index, 1 );

    // PIN State Query
    if( value == 0 ) {
        // RUIM_PRESENT                 1
        value = get_value( buf, &index, 1 );
        sprintf( buffer + strlen(buffer), "RESULT[0]: %d\n", value );

        // RUIM_BLOCK                   1
        value = get_value( buf, &index, 1 );
        sprintf( buffer + strlen(buffer), "RESULT[1]: %d\n", value );

        // RUIM_CHV                     1
        value = get_value( buf, &index, 1 );
        sprintf( buffer + strlen(buffer), "RESULT[2]: %d\n", value );

        // RUIM_CHV1                    1
        value = get_value( buf, &index, 1 );
        sprintf( buffer + strlen(buffer), "RESULT[3]: %d\n", value );
    }

    // PIN Verification
    else if( value == 1 ) {
        // RUIM_SUCCESS                 1
        value = get_value( buf, &index, 1 );
        sprintf( buffer + strlen(buffer), "RESULT[0]: %d\n", value );

        // RUIM_ERRCODE                 1
        value = get_value( buf, &index, 1 );
        sprintf( buffer + strlen(buffer), "RESULT[1]: %d\n", value );
    }

    // PIN RA NUM
    else if( value == 2 ) {
        // RUIM_PIN1                    1
        value = get_value( buf, &index, 1 );
        sprintf( buffer + strlen(buffer), "RESULT[0]: %d\n", value );

        // RUIM_PUK1                    1
        value = get_value( buf, &index, 1 );
        sprintf( buffer + strlen(buffer), "RESULT[1]: %d\n", value );
    }

    int buflen = strlen( buffer );
    if( buflen > LINE_DIM ) {
        printf( "small rxbuf size(%d) than %d\n", len, buflen );
        buflen = LINE_DIM;
        buffer[ buflen ] = '\0';
    }

    memcpy( (char*)buf, buffer, buflen );
    return buflen;
}

int dm_parse_ruim_chv( unsigned char *buf, unsigned int len )
{
    int  index = 0;
    int  value = 0;
    char buffer[ LINE_DIM + 2 ] = "";

    // CMD_CODE (0x8E)              1
    value = get_value( buf, &index, 1 );

    // SUB_CMD                      1
    value = get_value( buf, &index, 1 );

    // RUIM_SUCCESS                 1
    value = get_value( buf, &index, 1 );
    sprintf( buffer + strlen(buffer), "RESULT[0]: %d\n", value );

    // RUIM_ERRCODE                 1
    value = get_value( buf, &index, 1 );
    sprintf( buffer + strlen(buffer), "RESULT[1]: %d\n", value );

    int buflen = strlen( buffer );
    if( buflen > LINE_DIM ) {
        printf( "small rxbuf size(%d) than %d\n", len, buflen );
        buflen = LINE_DIM;
        buffer[ buflen ] = '\0';
    }

    memcpy( (char*)buf, buffer, buflen );
    return buflen;
}

int dm_parse_nv_read( unsigned char *buf, unsigned int len )
{
    int  index = 0;
    int  value = 0;
    char buffer[ LINE_DIM + 2 ] = "";

    if( len != 136 ) {
        sprintf( buf, "STATUS: Invalid length\n" );
        return strlen( buf );
    }

    // CMD_CODE (0x26)              1
    value = get_value( buf, &index, 1 );

    // NAM
    value = get_value( buf, &index, 1 );
    sprintf( buffer + strlen(buffer), "NAM: %d\n", value );

    // CODE
    value = get_value( buf, &index, 1 );
    sprintf( buffer + strlen(buffer), "CODE: %d\n", value );

    // DATA                         128
    int loop;
    sprintf( buffer + strlen(buffer), "DATA: ");
    for( loop = 0; loop < 128; loop++ ) {
        value = get_value( buf, &index, 1 );
        sprintf( buffer + strlen(buffer), "%02x", value );
    }
    sprintf( buffer + strlen(buffer), "\n");

    // STATUS                       2
    value = get_value( buf, &index, 2 );
    switch( value ) {
        case 0:
            sprintf( buffer + strlen(buffer), "STATUS: OK\n");
            break;
        case 1:
            sprintf( buffer + strlen(buffer), "STATUS: internal DMSS use\n");
            break;
        case 2:
            sprintf( buffer + strlen(buffer), "STATUS: Unregcognized command\n");
            break;
        case 3:
            sprintf( buffer + strlen(buffer), "STATUS: NV memory is full\n");
            break;
        case 4:
            sprintf( buffer + strlen(buffer), "STATUS: command failed\n");
            break;
        case 5:
            sprintf( buffer + strlen(buffer), "STATUS: Variable was not active\n");
            break;
        case 6:
            sprintf( buffer + strlen(buffer), "STATUS: Bad parameters\n");
            break;
        case 7:
            sprintf( buffer + strlen(buffer), "STATUS: Read-only\n");
            break;
        case 8:
            sprintf( buffer + strlen(buffer), "STATUS: Not defined\n");
            break;
        case 9:
            sprintf( buffer + strlen(buffer), "STATUS: No memory\n");
            break;
        case 10:
            sprintf( buffer + strlen(buffer), "STATUS: Internal use\n");
            break;
        default:
            sprintf( buffer + strlen(buffer), "STATUS: Unknown error\n");
            break;
    }

    int buflen = strlen( buffer );
    if( buflen > LINE_DIM ) {
        printf( "small rxbuf size(%d) than %d\n", len, buflen );
        buflen = LINE_DIM;
        buffer[ buflen ] = '\0';
    }

    memcpy( (char*)buf, buffer, buflen );
    return buflen;
}

int dm_parse_output( unsigned char *buf, unsigned int len )
{
    if( buf[ 0 ] == 0xCC ) {
        return dm_parse_state( buf, len );
    }
    else if( buf[ 0 ] == 0x0C ) {
        return dm_parse_status( buf, len );
    }
    else if( buf[ 0 ] == 0x40 ) {
        return dm_parse_pilot( buf, len );
    }
    else if( buf[ 0 ] == 0x8D ) {
        return dm_parse_ruim_pin( buf, len );
    }
    else if( buf[ 0 ] == 0x8E ) {
        return dm_parse_ruim_chv( buf, len );
    }
    else if( buf[ 0 ] == 0x26 ) {
        return dm_parse_nv_read( buf, len );
    }

    strcpy( buf, "null\n" );
    return( 0 );
}

void dm_notify( unsigned char *buf, unsigned int len )
{
    dm_parse_output( buf, len );

    if( dm_quiet > 0 ) {
        FILE *fp;

        fp = fopen( dm_file, "w+");
        if( fp == NULL ) {
            printf("Could not open file( %s )\n", dm_file);
            return;
        }

        fputs(buf, fp);
        fclose(fp);
    }
    else {
        printf( "%s\n", buf );
    }
}

/* Function to open a serial port */
int dm_open_port( void )
{
    int fd;
    struct termios tty;

    /* Open Port */
    fd = open( dm_port, O_RDWR | O_EXCL | O_NOCTTY | O_NONBLOCK );
    if( fd == -1 ) {
        printf( "Unable to open port( %s )\n", dm_port );
        return( -1 );
    }
    printf( "Successfully open port( %s )\n", dm_port );

    tcdrain( fd );
    tcflush( fd , TCIOFLUSH );

    if( tcgetattr(fd, &tty) < 0 ) {
        printf( "This helps in some cases\n" );
        tcsendbreak(fd, 0);
        sleep(1);
        if( tcgetattr(fd, &tty) < 0 ) {
            printf("The modem is busy.\n");
            close( fd );
            return( -1 );
        }
    }

    memset(&dm_tios,'\0',sizeof(dm_tios));

    dm_tios = tty;

    tty.c_cc[VMIN]  = 0; // nonblocking
    tty.c_cc[VTIME] = 0;
    tty.c_oflag     = 0;
    tty.c_lflag     = 0;

    tty.c_cflag &= ~(CSIZE | CSTOPB | PARENB);
    tty.c_cflag |= CS8 | CREAD;
    tty.c_cflag |= CLOCAL;            // ignore modem status lines
    tty.c_iflag  = IGNBRK | IGNPAR;
    tty.c_lflag &= ~ICANON;           // non-canonical mode
    tty.c_lflag &= ~(ECHO|ECHOE|ECHOK|ECHOKE);

    tty.c_cflag &= ~CRTSCTS;
    tty.c_iflag &= ~(IXON | IXOFF);

    cfsetospeed(&tty, B115200);
    cfsetispeed(&tty, B115200);

    tcdrain( fd );

    if( tcsetattr(fd, TCSANOW, &tty) < 0 ) {
        printf("The modem is busy.\n");
        close( fd );
        return( -1 );
    }

    return( fd );
}

/* Function to close a serial port */
int dm_close_port( int fd )
{
    tcflush( fd, TCIOFLUSH );

    if( tcsetattr( fd, TCSANOW, &dm_tios ) < 0 ) {
        close( fd );
        return( 0 );
    }

    close( fd );
    return( 1 );
}

int read_arguments(int argc, char **argv)
{
    int c, index = 0, nums = 0;
    if( argc == 1 ) return 0;

    dm_quiet = 0;
    dm_txlen = 0;
    memset( dm_txbuf, '\0', sizeof(dm_txbuf) );

    while( 1 ) {
        c = getopt_long( argc, argv, "hv:c:p:f:", long_options, &index );

        /* Detect the end of the options. */
        if( c == -1 ) break;

        nums++;
        switch( c ) {
            case 'c':
                if( strlen( optarg ) > sizeof(dm_txbuf) ) {
                    printf( "command length overmax %d bytes than %d bytes\n",
                            strlen(optarg), sizeof(dm_txbuf) );
                    dm_print_usage();
                    exit(0);
                }

                dm_txlen = strlen( optarg );
                memcpy( dm_txbuf, optarg, dm_txlen );
                printf( "CMD[%d]=%s\n", dm_txlen, dm_txbuf );
                break;

            case 'p':
                strcpy( dm_port, optarg );
                break;

            case 'f':
                dm_quiet = 1;
                strcpy( dm_file, optarg );
                break;

            case 'v':
                dm_print_version();
                exit(0);
                break;

            case 'h':
                dm_print_version();
                dm_print_usage();
                exit(0);
                break;

            default: // Unsupported - error message has already been printed
                dm_print_usage();
                exit(1);
        }
    }

    return nums;
}

int main(int argc, char **argv)
{
    int fd;

    if( ! read_arguments(argc, argv) ) {
        printf( "no arguments\n" );
        exit( 1 );
    }

    if( dm_txlen <= 0 ) {
        printf( "no command argument\n" );
        exit( 1 );
    }

    fd = dm_open_port();
    if( fd < 0 ) {
        exit( 1 );
    }

    dm_make_pkt( dm_txbuf, &dm_txlen );

    int txtry = MDM_TX_RETRY;
    do {
        struct timespec req = { 0, 10000000 }; /* 100 msec */

        if( dm_put_line( fd, dm_txbuf, dm_txlen ) < 0 ) {
            dm_close_port( fd );
            exit( 1 );
        }

        int rxtry = MDM_RX_RETRY;
        do {
            nanosleep(&req, NULL);

            dm_rxlen = dm_get_line( fd, dm_rxbuf, LINE_DIM );
            if( dm_rxlen > 0 ) {
                dm_notify( dm_rxbuf, dm_rxlen );
                dm_close_port( fd );
                exit( 0 );
            }
        } while( --rxtry > 0 );

        tcflush(fd, TCIOFLUSH);
        nanosleep(&req, NULL);
    } while( --txtry > 0 );

    if( dm_close_port( fd ) < 0 ) {
        printf( "Could not close properly a serial port\n" );
    }

    exit( 1 );
}
